import spacy
from sekg.text.extractor.domain_entity.word_util import WordUtil


class CompleteSubject:
    def __init__(self, nlp=None):
        if nlp is None:
            self.nlp = spacy.load('en', disable=["ner"])
        else:
            self.nlp = nlp

    def complete_subject_by_name_for_doc(self, full_doc, sentence_belong_name, doc=None):
        if doc is not None:
            test_doc = doc
        else:
            test_doc = self.nlp(full_doc)
        result = []
        for sent in test_doc.sents:
            result.append(self.complete_subject_by_name_for_sentence(sent.text, sentence_belong_name, sent))
        return " ".join(result), test_doc

    def how_many_upper(self, input_text):
        """
        有多少个大写字符
        :param input_text:
        :return:
        """
        n = 0
        for l, r in zip(input_text.lower(), input_text):
            if l != r:
                n += 1
        return n

    def complete_subject_by_name_for_sentence(self, text, sentence_belong_name, sent):
        """
        根据文档所属的全限定名，补全
        先分析
        :param sent:
        :param text:
        :param sentence_belong_name:
        :return:补全后的句子
        """
        lower_text = str(text).lower()
        if self.check_has_main_subject(sent) or lower_text.lower().find(" this ") >= 0:
            return text
        else:
            for i, doc in enumerate(sent):
                if i == 0:
                    if WordUtil.couldBeVerb(doc.text):
                        f = str(text[0]).lower()
                        sentence = sentence_belong_name + " " +  f + text[1:]
                        return sentence
                    else:
                        if self.how_many_upper(doc.text) > 1:
                            sentence = sentence_belong_name + " is " + text
                            return sentence
                        else:
                            f = str(text[0]).lower()
                            sentence = sentence_belong_name + " is " + f + text[1:]
                            return sentence

    def check_has_main_subject(self, sent):
        # doc = self.nlp(text)
        for i, d in enumerate(sent):
            if i == 0 and d.dep_ == 'auxpass':
                return False
            if "nsubj" == d.dep_ and d.text != "that":
                return True
        return False


# if __name__ == '__main__':
#     cs = CompleteSubject()
#     s = """A mutable sequence of characters. This class provides an API compatible with StringBuffer, but with no guarantee of synchronization. This class is designed for use as a drop-in replacement for StringBuffer in places where the string buffer was being used by a single thread (as is generally the case). Where possible, it is recommended that this class be used in preference to StringBuffer as it will be faster under most implementations.
# The principal operations on a StringBuilder are the append and insert methods, which are overloaded so as to accept data of any type. Each effectively converts a given datum to a string and then appends or inserts the characters of that string to the string builder. The append method always adds these characters at the end of the builder; the insert method adds the characters at a specified point.
#
# For example, if z refers to a string builder object whose current contents are "start", then the method call z.append("le") would cause the string builder to contain "startle", whereas z.insert(4, "le") would alter the string builder to contain "starlet".
#
# In general, if sb refers to an instance of a StringBuilder, then sb.append(x) has the same effect as sb.insert(sb.length(), x).
#
# Every string builder has a capacity. As long as the length of the character sequence contained in the string builder does not exceed the capacity, it is not necessary to allocate a new internal buffer. If the internal buffer overflows, it is automatically made larger.
#
# Instances of StringBuilder are not safe for use by multiple threads. If such synchronization is required then it is recommended that StringBuffer be used."""
#     s = 'obtain a reference to a bootstrap remote object registry on a particular host (including the local host), or to create a remote object registry that accepts calls on a specific port.'
#     # java.lang.StringBuffer is A thread-safe, mutable sequence of characters.
#     after = cs.complete_subject_by_name_for_doc(s, "java.rmi.registry.LocateRegistry")
#     # after = cs.complete_subject_by_name_for_doc(s, "java.lang.StringBuffer")
#     print(after)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sekg.term import abbr_rules


class AbbrExtractor:
    def __init__(self, **cfg):
        rule_names = cfg.get("rules", ["Rule4Bracket", "Rule4Colon"])
        self.RULES = []
        for r in rule_names:
            if not hasattr(abbr_rules, r):
                print("Has no rule:", r)
                continue
            self.RULES.append(getattr(abbr_rules, r)())

    def extract(self, sentence):
        terms = set()
        pairs = set()
        for rule in self.RULES:
            _terms, _pairs = rule.extract(sentence)
            terms.update(_terms)
            pairs.update(_pairs)
        return terms, pairs

    def process(self, sentences):
        terms = set()
        pairs = set()
        for sent in sentences:
            _terms, _pairs = self.extract(sent)
            terms.update(_terms)
            pairs.update(_pairs)

        return terms, pairs



#!/usr/bin/python
# -*- coding: UTF-8 -*-
import re

import gensim
from bs4 import BeautifulSoup


class CodeTextPreprocessor:
    pattern = re.compile(r'\s+')
    CODE_FRAGMENT_MARK = "-CODE-"

    def __clean_format(self, text):
        """
        clean text format for text extract from html
        :param text:
        :return:
        """
        return re.sub(self.pattern, " ", text.replace('\n', ' ').replace(u'\u00a0', " "))

    def clean_html_text(self, html_text):
        if html_text is None or len(html_text) == 0:
            return ""

        soup = BeautifulSoup(html_text, "lxml")
        codeTags = soup.find_all(name=["pre", 'blockquote'])

        for tag in codeTags:
            tag.string = " " + self.CODE_FRAGMENT_MARK + " . \n"

        ## todo: this is for <li> <ul>,
        # list_groups = soup.find_all(name=["ol", "ul"])
        # for list_group in list_groups:
        #     list_items = list_group.find_all("li")
        #     num = 1
        #     for item in list_items:
        #         item.string = "{0}.{1}\n.".format(str(num), item.string)
        #         num = num + 1
        #
        # # todo: the sentence may lack of Punctuation mark in every <p> tag end. it will be

        cleanText = soup.get_text()
        decode_clean_text = gensim.utils.decode_htmlentities(cleanText)
        ## todo: test the code
        return self.__clean_format(decode_clean_text)



import pickle
import re
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from sekg.util.annotation import conn_try_again
from sekg.util.url_util import URLUtil
# re connect
from sekg.wiki.WikiDataItem import WikiDataItem
from sekg.wiki.search_domain_wiki.wiki_info_model import WikiInfo


class WikiTool:
    def __init__(self, domain_name_set_to_search, use_proxies=False, max_candidate_num=3):
        self.title_2_wiki = dict()
        self.wiki_data_id_2_wikidata = dict()
        self.domain_name_set_to_search = domain_name_set_to_search
        self.use_proxies = use_proxies
        self.max_candidate_num = max_candidate_num

    def save(self, save_path):

        with open(str(Path(save_path) / "title_2_wiki_dict"), 'wb') as f:
            pickle.dump(self.title_2_wiki, f, pickle.HIGHEST_PROTOCOL)

        with open(str(Path(save_path) / "wiki_data_id_2_wikidata_dict"), 'wb') as f:
            pickle.dump(self.wiki_data_id_2_wikidata, f, pickle.HIGHEST_PROTOCOL)

    def load(self, save_path):
        with open(str(Path(save_path) / "title_2_wiki_dict"), 'rb') as f:
            self.title_2_wiki = pickle.load(f)
        with open(str(Path(save_path) / "wiki_data_id_2_wikidata_dict"), 'rb') as f:
            self.wiki_data_id_2_wikidata = pickle.load(f)

    def get_wiki_id(self, wd_url):
        return (str(wd_url).split('/'))[-1]

    def find(self, title):
        if title in self.title_2_wiki:
            return self.title_2_wiki[title]
        else:
            return self.search_title(title)

    def search_title(self, title):
        wiki_info_list = []
        wiki_id_list, wikipedia_url_list, wikipedia_title_list, wd_item_list = (self.get_wd_by_title(title))

        for wiki_id, wikipedia_url, wikipedia_title, wd_item in zip(wiki_id_list, wikipedia_url_list,
                                                                    wikipedia_title_list, wd_item_list):
            wiki_info = WikiInfo(wiki_data_id=wiki_id, wikipedia_url=wikipedia_url, wikipedia_title=wikipedia_title,
                                 wiki_data_item=wd_item)
            wiki_info_list.append(wiki_info)

        self.title_2_wiki[title] = wiki_info_list
        return wiki_info_list

    def start_search(self):
        for title in self.domain_name_set_to_search:
            self.search_title(title)

    @conn_try_again
    def get_wd_by_title(self, title):
        wiki_data_id_list = []
        wikipedia_url_list = []
        wikipedia_title_list = []
        if self.use_proxies:
            proxies = {"http": "127.0.0.1:1080", "https": "127.0.0.1:1080"}
        else:
            proxies = None
        try:
            base_url = "https://en.wikipedia.org"
            url = base_url + "/wiki/" + title
            m = requests.get(url, proxies=proxies)
            if m.status_code == 200:
                soup = BeautifulSoup(m.content, "lxml")
                wikidata_node_list = soup.select('#t-wikibase')
                if wikidata_node_list:
                    wikidata_node = wikidata_node_list[0]
                    for link in wikidata_node.findAll('a', attrs={'href': re.compile("^https://")}):
                        wd_href = link.get('href')
                        wd_id = self.get_wiki_id(wd_href)
                        wikipedia_title = URLUtil.parse_url_to_title(m.url)

                        wiki_data_id_list.append(wd_id)
                        wikipedia_url_list.append(m.url)
                        wikipedia_title_list.append(wikipedia_title)
                else:
                    print("wikidata_node_list is empty")
            else:
                new_title = title.replace(" ", "+")
                url = "https://en.wikipedia.org/w/index.php?search=" + new_title + "&title=Special%3ASearch&go=Go"
                m = requests.get(url, proxies=proxies)
                if m.status_code == 200:
                    soup = BeautifulSoup(m.content, "lxml")
                    candidate_list = soup.select('.mw-search-result-heading')
                    for i, candidate in enumerate(candidate_list):
                        if i == self.max_candidate_num - 1:
                            break
                        for link in candidate.findAll('a', attrs={'href': re.compile("^/wiki/")}):
                            wd_href = link.get('href')
                            redirect_url = base_url + wd_href
                            redirect_m = requests.get(redirect_url, proxies=proxies)
                            if redirect_m.status_code == 200:
                                soup = BeautifulSoup(redirect_m.content, "lxml")
                                wikidata_node_list = soup.select('#t-wikibase')
                                if wikidata_node_list:
                                    wikidata_node = wikidata_node_list[0]
                                    for link in wikidata_node.findAll('a', attrs={'href': re.compile("^https://")}):
                                        wd_href = link.get('href')
                                        wd_id = self.get_wiki_id(wd_href)
                                        if len(wd_id) > 1 and wd_id[0] == "Q":
                                            wiki_data_id_list.append(wd_id)
                                            wikipedia_url_list.append(redirect_m.url)
                                            wikipedia_title_list.append(URLUtil.parse_url_to_title(redirect_m.url))
                                            break
                else:
                    print("fail to find")
            wd_item_list = []
            for wiki_data_id in wiki_data_id_list:
                if wiki_data_id not in self.wiki_data_id_2_wikidata:
                    wd_item = WikiDataItem(wiki_data_id)
                    if wd_item:
                        self.wiki_data_id_2_wikidata[wiki_data_id] = wd_item
                else:
                    wd_item = self.wiki_data_id_2_wikidata[wiki_data_id]
                wd_item_list.append(wd_item)

            return wiki_data_id_list, wikipedia_url_list, wikipedia_title_list, wd_item_list
        except Exception as e:
            print(e)



from unittest import TestCase

from sekg.mysql.factory import MysqlSessionFactory


class TestMysqlSessionFactory(TestCase):
    def test_create_mysql_engine_by_server_name(self):
        factory = MysqlSessionFactory("mysql_config.json")

        engine = factory.create_mysql_engine_by_server_name("87RootServer")
        self.assertIsNotNone(engine)

        engine = factory.create_mysql_engine_by_server_name("DysD3Wrong")
        self.assertIsNone(engine)

    def test_create_mysql_engine_by_server_id(self):
        factory = MysqlSessionFactory("mysql_config.json")
        engine = factory.create_mysql_engine_by_server_id(1)
        self.assertIsNotNone(engine)

        engine3 = factory.create_mysql_engine_by_server_id(3)
        self.assertIsNone(engine3)

    def test_create_mysql_session_by_server_name(self):
        factory = MysqlSessionFactory("mysql_config.json")

        session = factory.create_mysql_session_by_server_name("87RootServer", database="testsekg")
        self.assertIsNotNone(session)

        session = factory.create_mysql_session_by_server_name("DysD3Wrong", database="testsekg")
        self.assertIsNone(session)

    def test_create_mysql_session_by_server_id(self):
        factory = MysqlSessionFactory("mysql_config.json")
        session = factory.create_mysql_session_by_server_id(1, database="testsekg")
        self.assertIsNotNone(session)

        engine3 = factory.create_mysql_session_by_server_id(3, database="testsekg")
        self.assertIsNone(engine3)

    def test_create_all_databases(self):
        factory = MysqlSessionFactory("mysql_config.json")
        factory.create_all_databases(server_name="87RootServer")



class WikiInfo:
    def __init__(self, wiki_data_id, wikipedia_url, wikipedia_title, wiki_data_item):
        self.wiki_data_id = wiki_data_id
        self.wikipedia_url = wikipedia_url
        self.wikipedia_title = wikipedia_title
        self.wiki_data_item = wiki_data_item

    def __repr__(self):
        return "<WikiInfo wd_id=%r title=%r>" % (self.wiki_data_id, self.wikipedia_title)



from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.preprocessor.base import Preprocessor

import re
from itertools import combinations
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer

class NGramPreprocessor(Preprocessor):
    LENGTH_PRIORITY = 0
    ORDER_PRIORITY = 1

    def __init__(self, n, kg_name_searcher):
        self.n = n
        self.lemmatizer = WordNetLemmatizer()
        self.name_searcher = kg_name_searcher

    def get_n_gram_list(self, query, split_type=LENGTH_PRIORITY):
        entities = []
        words = re.split(r"\"|'", query)
        if len(words) % 2 == 0:
            words.pop(len(words) - 1)
        for index, np in enumerate(words):
            if index % 2 == 1 and len(str(np)) > 0:
                lemma_token = self.lemmatizer.lemmatize(np)
                # 将得到的特殊词汇进行替换 替换为NP
                query = query.replace(str("\"" + np + "\""), "Entity_NP")
                entities.append(lemma_token)
        del words
        paragraph = re.split("[,|.|?|!|:|;]", query)
        # 处理引号问题
        for p in paragraph:
            words = [self.lemmatizer.lemmatize(w) for w in re.split("[\s|/|+|\\|(|)]", p.lower()) if len(w) > 0]
            print(words)
            # words = [w for w in words if w not in stopwords.words('english')]
            no_word = []
            delete_index = []
            w_size = self.n
            if split_type == self.ORDER_PRIORITY:
                index = 0
                while index < len(words):
                    t_tag = min(index + w_size, len(words) - index)
                    while t_tag >= 1:
                        if len(self.name_searcher.search_by_full_name(" ".join(words[index:index + t_tag]))) > 0:
                            entities.append(" ".join(words[index:index + t_tag]))
                            index += t_tag
                            break
                        else:
                            t_tag -= 1
                    if t_tag == 0:
                        no_word.append(words[index])
                        index += 1
            elif split_type == self.LENGTH_PRIORITY:
                index = 0
                while index < len(words):
                    t_tag = min(index + w_size, len(words) - index)
                    while t_tag >= 1:
                        for item in list(combinations(words[index:index + w_size], t_tag)):
                            if len(self.name_searcher.search_by_full_name(" ".join(item))) > 0:
                                if " ".join(list(item)) not in entities:
                                    entities.append(" ".join(list(item)))
                        t_tag -= 1
                    index += 1
        new_entities = []
        for key in entities:
            tag = False
            for item in entities:
                if key != item and key in item:
                    tag = True
            if not tag and key not in stopwords.words('english'):
                new_entities.append(key)
                new_entities = [w for w in new_entities if w.isalnum() and not w.isnumeric()]
        return new_entities


if __name__ == "__main__":
    name_searcher = KGNameSearcher.load(
        "E:\PycharmProjects\KGFeatureLocation\output\graph\jedite-4.3\\NameSearcher\jedite-4.3.v4.2.namesearcher"
    )
    preprocessor = NGramPreprocessor(4, kg_name_searcher=name_searcher)
    query_str = [
        "folding bug, text is in a black hole",
        "Size of file open/save dialogs incorrect on dual displays",
        "Incorrect Python code indentation",
        "Search expression",
        "Autosave turned off for Untitled Documents",
        "Keyboard navigation through menubar inserts char into buffer",
        "Add filter to recent files menu",
        "File Open Dialog: can't handle full pathnames",
        "Provide a way of reloading changed buffers without prompting",
        "Rectangular Selection and Dragging",
        "Filter Bug in File System Browser",
        "Auto-indent breaks if previous line is a single line indent",
        "set \"Search for:\" to \"||\" check \"Regular expressions\" click \"Replace All\" version : jEdit 4.3pre8",
        "Wrong conversion to UTF-16 (with BOM)",
        "caret/display position w/ multiple views open",
    ]
    for query in query_str:
        print(preprocessor.get_n_gram_list(query, preprocessor.LENGTH_PRIORITY))





from sekg.graph.exporter.graph_data import GraphData
from sekg.util.vector_util import VectorUtil


import math
import numpy as np
from sekg.graph.exporter.graph_data import GraphData
from sekg.graph.exporter.graph_data import GraphIndexCollection
from sekg.graph.metadata_accessor import MetadataGraphAccessor
import networkx as nx
from networkx import MultiGraph

from gensim.utils import SaveLoad

"""
this package is used to get a weight graph
"""


class WeightGraphData(SaveLoad):
    """
    the store of a weight undirected graph data.

    each node is represent as a dict of node info named 'node_json',
    Example Format for 'node_json':

     {
        "id": 1,
        "properties": {"name":"bob","age":1},
        "labels": ["entity","man"]
    }

    """

    DEFAULT_KEY_NODE_ID = "id"  # the key name for the node id, every node must have it.
    DEFAULT_KEY_NODE_PROPERTIES = "properties"  # the key name for the node properties, every node must have it.
    DEFAULT_KEY_NODE_LABELS = "labels"  # the key name for the node labels, every node must have it.

    DEFAULT_KEYS = [DEFAULT_KEY_NODE_ID, DEFAULT_KEY_NODE_PROPERTIES, DEFAULT_KEY_NODE_LABELS]
    UNASSIGNED_NODE_ID = -1  # a node without a id specify, a newly created node, its id is -1

    DEFAULT_KEY_RELATION_START_ID = "startId"
    DEFAULT_KEY_RELATION_TYPE = "relationType"
    DEFAULT_KEY_RELATION_END_ID = "endId"

    DEFAULT_KEY_PROPERTY_QUALIFIED_NAME = "qualified_name"
    DEFAULT_KEY_PROPERTY_ALIAS = "alias"

    def __init__(self):
        self.graph = MultiGraph()
        self.max_node_id = 0
        self.label_to_ids_map = {}
        self.index_collection = GraphIndexCollection()
        self.relation_type_weight_dict = {}
        self.node_weight_dict = {}
        self.r_weights = {}

    '''
    init the undirected graph from directed graph and get the node weight dict
    :param graph_data: MulDiGraph
    '''
    def init_from_directed_graph(self, graph_data, weight="weight"):
        if not isinstance(graph_data, GraphData):
            # todo alert an Error info
            return
        print("Transfer undirected graph to weight graph data")
        self.label_to_ids_map = graph_data.label_to_ids_map
        computer = WeightGraphDataComputer(graph_data)
        self.graph = computer.generate_weight_graph(weight_type=weight)

    def get_relation_num(self):
        return len(self.graph.edges)

    def get_node_ids(self):
        return set(self.graph.nodes)

    def get_node_label_by_id(self, node_id):
        pass


    # todo: fix the weight graphdata training
    # def precompute_weight(self):
    #     relation_count = self.graph_data.get_relation_num()
    #     relation_type_count = {}
    #
    #     node_degree_dict = {}
    #
    #     for node_id in self.graph_data.get_node_ids():
    #         # get the out degree of a node
    #         out_relation_list = self.graph_data.get_all_out_relations(node_id=node_id)
    #         # get the in degree of a node
    #         in_relation_list = self.graph_data.get_all_in_relations(node_id=node_id)
    #
    #         # get the degree of a node
    #         node_degree_dict[node_id] = len(out_relation_list) + len(in_relation_list)
    #
    #         for start_id, relation_type, end_id in out_relation_list:
    #             if relation_type not in relation_type_count.keys():
    #                 relation_type_count[relation_type] = 0
    #             relation_type_count[relation_type] = relation_type_count[relation_type] + 1
    #
    #     self.relation_type_weight_dict = VectorUtil.compute_idf_weight_dict(total_num=relation_count,
    #                                                                         number_dict=relation_type_count)
    #     print(self.relation_type_weight_dict)
    #     self.node_weight_dict = VectorUtil.compute_idf_weight_dict(total_num=relation_count * 2,
    #                                                                number_dict=node_degree_dict)

    def get_node_weight(self, node_id):
        if node_id not in self.node_weight_dict.keys():
            return 1
        return self.node_weight_dict[node_id]

    def get_relation_type_weight(self, relation_type):
        if relation_type not in self.relation_type_weight_dict.keys():
            return 1
        return self.relation_type_weight_dict[relation_type]

    def get_relation_tuple_weight(self, start_node_id, end_node_id, relation_name):
        start_w = self.get_node_weight(start_node_id)
        end_w = self.get_node_weight(end_node_id)
        r_w = self.get_relation_type_weight(relation_name)

        return start_w * end_w * r_w

    def get_all_in_relations(self, node_id):
        if node_id not in self.graph.nodes:
            return set()
        return {(r[0], r[2], r[1]) for r in self.graph.edges(node_id, keys=True)}


class WeightGraphDataComputer:
    """
    this class is to get a weight graph Data"""

    def __init__(self, graph_data):
        if isinstance(graph_data, GraphData):
            self.graph_data = graph_data
        else:
            self.graph_data = None

        self.undirected_graph = None
        self.max_w_r_type_dict = {}
        self.node_id_to_weight_map = {}
        self.r_weights = {}

    '''
    weight_type == "weight" or weight_type == "distance"
    :param weight_type: 
    :return: 
    '''
    def generate_weight_graph(self, weight_type="weight"):
        r_weights, max_w_r_type_dict = self.compute_r_weight(self.graph_data)
        self.max_w_r_type_dict = max_w_r_type_dict
        self.r_weights = r_weights

        graph_data: GraphData = self.graph_data
        undirected_graph = graph_data.graph.to_undirected()
        node_id_to_weight_map = self.compute_node_weight(undirected_graph)
        self.node_id_to_weight_map = node_id_to_weight_map

        r_list = []
        r_weight_list = []
        r_type_list = []
        for start_id, end_id, relation_type in undirected_graph.edges(keys=True):
            r_list.append((start_id, end_id))
            r_type_list.append(relation_type)
            start_w = node_id_to_weight_map[start_id]
            end_w = node_id_to_weight_map[end_id]
            r_weight = r_weights[(start_id, end_id)]
            weight = (start_w + end_w) * r_weight

            r_weight_list.append(weight)

        np_r_weight = np.array(r_weight_list)
        w_min = np_r_weight.min()
        w_max = np_r_weight.max()
        normal_r_weight = (np_r_weight - w_min) / (w_max - w_min)
        normal_r_weight = np.abs(1 - normal_r_weight)
        undirected_graph = nx.MultiGraph()

        for (start_id, end_id), weight, normal_w, relation_type in zip(r_list, r_weight_list, normal_r_weight, r_type_list):
            if weight_type == "weight":
                undirected_graph.add_edge(start_id, end_id, weight=1 - normal_w, key=relation_type)
            else:
                undirected_graph.add_edge(start_id, end_id, weight=normal_w, key=relation_type)

        self.undirected_graph = undirected_graph
        return undirected_graph

    @staticmethod
    def compute_r_weight(graph_data):
        relation_count = graph_data.get_relation_num()
        print("relation count=%d" % relation_count)
        relation_statics = {}
        # get the num of relation type and count for every node
        node_id_to_relation_type_count = {}
        r_to_relation_types_map = {}

        def _count_node_r_weight(node_id, relation_type):
            if node_id not in node_id_to_relation_type_count:
                node_id_to_relation_type_count[node_id] = dict()

            if relation_type not in node_id_to_relation_type_count[node_id]:
                node_id_to_relation_type_count[node_id][relation_type] = 0
            node_id_to_relation_type_count[node_id][relation_type] += 1

        for start_id, relation_type, end_id in graph_data.get_relation_pairs_with_type():
            if relation_type.startswith("operation_"):
                relation_type = "operation"
            if relation_type.startswith("mention "):
                relation_type = "mention"

            _count_node_r_weight(start_id, relation_type)
            _count_node_r_weight(end_id, relation_type)
            if relation_type not in relation_statics:
                relation_statics[relation_type] = 0

            relation_statics[relation_type] += 1
            if (start_id, end_id) not in r_to_relation_types_map:
                r_to_relation_types_map[(start_id, end_id)] = set([])
            if (end_id, start_id) not in r_to_relation_types_map:
                r_to_relation_types_map[(end_id, start_id)] = set([])
            r_to_relation_types_map[(start_id, end_id)].add(relation_type)
            r_to_relation_types_map[(end_id, start_id)].add(relation_type)

        r_weights = {}
        for node_id in node_id_to_relation_type_count.keys():
            for r in node_id_to_relation_type_count[node_id].keys():
                node_id_to_relation_type_count[node_id][r] = 1.0 / (math.log10(
                    node_id_to_relation_type_count[node_id][r]) + 1)

        max_w_r_type_dict = {}
        for (start_id, end_id), relation_types in r_to_relation_types_map.items():
            r_weights[(start_id, end_id)] = 0
            for r in relation_types:
                w = 2 / (1/node_id_to_relation_type_count[start_id][r] + 1/node_id_to_relation_type_count[end_id][r])
                if w > r_weights[(start_id, end_id)]:
                    r_weights[(start_id, end_id)] = w
                    max_w_r_type_dict[(start_id, end_id)] = r
        return r_weights, max_w_r_type_dict

    @staticmethod
    def compute_node_weight(undirected_graph):
        node_id_list = []
        node_degree_list = []
        edge_num = undirected_graph.number_of_edges()
        node_id_to_weight_map = {}
        for node_id in undirected_graph.nodes():
            node_id_list.append(node_id)
            node_degree_list.append(undirected_graph.degree(node_id))
        node_weight_list = np.log10((edge_num + 1) / (np.array(node_degree_list) + 1))
        for node_id, weight in zip(node_id_list, node_weight_list):
            node_id_to_weight_map[node_id] = weight
        return node_id_to_weight_map

    def save(self, edge_path):
        # print([e for e in self.undirected_graph.edges.data('weight', default=1)])
        with open(edge_path, "w+") as f:
            f.writelines([
                " ".join([str(item) for item in e]) for e in self.undirected_graph.edges.data('weight', default=1)
            ])

    def get_distance(self, start_id, end_id):
        if start_id not in self.undirected_graph.nodes:
            return -1
        elif end_id not in self.undirected_graph.nodes:
            return -1
        for adj in self.undirected_graph.edges.data('weight', default=1):
            if (start_id == adj[0] and end_id == adj[1]) or (start_id == adj[1] and end_id == adj[0]):
                return adj[2]
        return -1

    def get_weight(self, start_id, end_id):
        if start_id not in self.undirected_graph.nodes:
            return -1
        elif end_id not in self.undirected_graph.nodes:
            return -1
        for adj in self.undirected_graph.edges.data('weight', default=1):
            if (start_id == adj[0] and end_id == adj[1]) or (start_id == adj[1] and end_id == adj[0]):
                return 1 - adj[2]
        return -1


if __name__ == "__main__":
    graph_path = "E:\PycharmProjects\KGFeatureLocation\output\graph\jedite-4.3\jedite-4.3.v3.graph"
    c = WeightGraphDataComputer(GraphData.load(graph_path))
    c.generate_weight_graph(weight_type="weight")
    print(c.get_weight(40153, 40154))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import networkx as nx
from gensim.models import Word2Vec
from gensim.models.word2vec import LineSentence

from sekg.graph.exporter.weight_graph_data import WeightGraphData
from sekg.model.node2vec import node2vec


from sekg.graph.exporter.graph_data import GraphData
from pathlib import Path


class GraphNode2VecTrainer:
    def __init__(self, graph_data, directed=False, weight="weight"):
        self.node_num = graph_data.get_node_num()
        self.nx_G_instance = None
        if directed:
            self.init_weight_graph(graph_data, weight=weight)
        else:
            self.graph_data = graph_data
            self.init_unweight_graph()

    def init_unweight_graph(self):
        node_ids = self.graph_data.get_node_ids()
        relation_pairs = self.graph_data.get_relation_pairs()
        print("node num=%d" % self.node_num)
        print("relation num=%d" % len(relation_pairs))
        G = nx.DiGraph()
        G.add_nodes_from(node_ids)
        G.add_edges_from(relation_pairs, weight=1.0)
        # todo: a relation weight support
        self.nx_G_instance = G
        print("init graph trainer by unweight relations")

    def init_weight_graph(self, graph_data, weight="weight"):
        '''
        INIT WEIGHT GRAPH: init weight graph by unweight graph and edge weight list
        :param weight_list: [(id1, id2, weight)]
        :return:
        '''
        # First Step: Compute the weight of a
        node_ids = graph_data.get_node_ids()
        relation_pairs_with_type = graph_data.get_relation_pairs_with_type()
        print("node num=%d" % self.node_num)
        print("relation num=%d" % len(relation_pairs_with_type))

        print(" pre-compute weight start")

        self.graph_data = WeightGraphData()
        self.graph_data.init_from_directed_graph(graph_data=graph_data, weight=weight)
        print("pre-compute weight end")

        weight_relation_tuples = []
        for (start_id, relation_type, end_id) in relation_pairs_with_type:
            weight = self.graph_data.get_relation_tuple_weight(start_node_id=start_id, relation_name=relation_type,
                                                               end_node_id=end_id)
            weight_relation_tuples.append((start_id, end_id, weight))

        G = nx.DiGraph()
        G.add_nodes_from(node_ids)
        G.add_weighted_edges_from(weight_relation_tuples)
        self.nx_G_instance = G
        print("init graph trainer by weighted relations(tf-idf tuple)")

    def generate_random_path(self, rw_path_store_path, directed=False, p=1, q=1, num_walks=10, walk_length=80,
                             ):
        print("start generate graph random path")

        G = node2vec.Graph(self.nx_G_instance, directed, p, q)
        G.preprocess_transition_probs()
        walks = G.simulate_walks(num_walks, walk_length)
        # todo: may be load all into memory has problem?
        with open(rw_path_store_path, 'w') as write_f:
            for walk in walks:
                path_str = " ".join([str(item) for item in walk])
                write_f.writelines('%s\n' % (path_str))

        print("complete generate graph random path")

    @staticmethod
    def train(rw_path_store_path, model_path, dimensions=100,
              workers=4):
        """
        train the graph vector from rw_path
        :param rw_path_store_path: the random walk for one graph
        :param model_path: the output word2vec model path
        :param dimensions: the dimensions of word2vec
        :param workers: the num of pipeline training
        :return:
        """
        print("save graph2vec training")

        # Learn embeddings by optimizing the Skipgram objective using SGD.
        w2v = Word2Vec(LineSentence(rw_path_store_path), size=dimensions, min_count=0, sg=1, workers=workers)
        w2v.wv.save(model_path)
        print("save graph2vec to %s" % model_path)
        return w2v



from unittest import TestCase

from sekg.ir.preprocessor.base import SimplePreprocessor


class TestSimplePreprocessor(TestCase):
    def test_remove_special_char(self):
        test_case_list = [
            (
                "This is a ArrayList, it contains all of the classes for creating user interfaces and for painting graphics and images."

                ,
                "This is a ArrayList it contains all of the classes for creating user interfaces and for painting graphics and images"),

            ("how to get File's MD5 checksum",

             "how to get File s MD5 checksum")

        ]
        preprocessor = SimplePreprocessor()
        for old_str, new_str in test_case_list:
            team = preprocessor.remove_special_char(old_str)
            self.assertEqual(team, new_str)

    def test_clean(self):

        test_case_list = [
            (
                "This is a ArrayList, it contains all of the classes for creating user interfaces and for painting graphics and images."

                ,
                "This is a ArrayList it contains all of the classes for creating user interfaces and for painting graphics and images"),

            ("how to get File's MD5 checksum",

             "how to get File s MD5 checksum")

        ]
        preprocessor = SimplePreprocessor()
        for old_str, new_str in test_case_list:
            self.assertEqual(new_str.lower().split(), preprocessor.clean(old_str))



#!/usr/bin/python
# -*- coding: UTF-8 -*-
from pathlib import Path

import gensim

from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.models.base import DocumentSimModel


# todo: the index file will has problem when the model are copy to new path
class TFIDFModel(DocumentSimModel):
    """
    This class is used the TF-IDF model to compute the document similarity score.
    The implementation from TFIDF is from gensim.
    """
    __tfidf_model_name__ = "tfidf.model"
    __corpus_name__ = "corpus.mm"
    __dictionary_name__ = "corpus.dict"
    # the index file is created for speeding up the similarity score computing
    __tfidf_sim_index__ = "TFIDF.sim.index"
    # the name for the PreprocessMultiFieldDocumentCollection. store
    __doc_collection_name__ = "doc.pre.collection"

    def __init__(self, name, model_dir_path, **config):

        """
        init the lsi model with
        :param model_dir_path:
        """
        super().__init__(name, model_dir_path, **config)

        self.tfidf_model = None
        self.corpus = None
        self.dict = None
        self.similarity_tfidf_index = None
        self.__init_sub_model_path__()

        self.preprocessor = None
        self.preprocess_doc_collection = None

    def __init_sub_model_path__(self):
        if self.model_dir_path is None:
            return
        # init the paths of models
        model_dir = Path(self.model_dir_path)
        self.tfidf_model_path = str(model_dir / self.__tfidf_model_name__)
        self.corpus_path = str(model_dir / self.__corpus_name__)
        self.dictionary_path = str(model_dir / self.__dictionary_name__)
        self.entity_collection_path = str(model_dir / self.__doc_collection_name__)

        index_dir = model_dir / "index"
        self.sim_index_dir = str(index_dir)
        self.sim_index_path = str(index_dir / self.__tfidf_sim_index__)

        model_dir.mkdir(parents=True, exist_ok=True)
        index_dir.mkdir(parents=True, exist_ok=True)

    def init_model_as_submodel(self):
        """
        init the model
        :return:
        """
        self.__init_sub_model_path__()
        print("loading the TFIDF models")
        self.tfidf_model = gensim.models.TfidfModel.load(self.tfidf_model_path)
        self.dict = gensim.corpora.Dictionary.load(self.dictionary_path)
        self.similarity_tfidf_index = gensim.similarities.Similarity.load(self.sim_index_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)
        print("load TFIDF models done")
        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.set_preprocessor(preprocess_doc_collection.get_preprocessor())
        self.__init_index()

    def init_model(self):
        """
        init the model
        :return:
        """
        self.__init_sub_model_path__()
        print("loading the TFIDF models")
        self.tfidf_model = gensim.models.TfidfModel.load(self.tfidf_model_path)
        self.corpus = gensim.corpora.MmCorpus(self.corpus_path)
        self.dict = gensim.corpora.Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)
        print("load TFIDF models done")
        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.__init_document_collection(preprocess_doc_collection)
        self.__init_index()

    def __init_document_collection(self, preprocess_doc_collection):
        self.preprocess_doc_collection = preprocess_doc_collection
        self.preprocessor = self.preprocess_doc_collection.get_preprocessor()

    def train_from_document_collection(self, preprocess_document_collection,
                                       pretrain_tfidf_path=None,
                                       pretrain_dictionary_path=None):
        print("start training")
        self.__init_document_collection(preprocess_document_collection)
        corpus_clean_text = []
        preprocess_multi_field_doc_list = preprocess_document_collection.get_all_preprocess_document_list()

        for docno, multi_field_doc in enumerate(preprocess_multi_field_doc_list):
            corpus_clean_text.append(multi_field_doc.get_document_text_words())
            # print("process doc %d" % docno)

        print("corpus len=%d" % len(corpus_clean_text))

        if pretrain_dictionary_path is not None:
            print("pretrain pretrain dictionary path is given, loading")
            self.dict = gensim.corpora.Dictionary.load(pretrain_dictionary_path)
        else:
            print("Dictionary init...")
            self.dict = gensim.corpora.Dictionary(corpus_clean_text)
            print("Dictionary init complete")

        self.corpus = [self.dict.doc2bow(text) for text in corpus_clean_text]

        if pretrain_tfidf_path is not None:
            print("pretrain tfidf path is given, loading")
            self.tfidf_model = gensim.models.TfidfModel.load(pretrain_tfidf_path)

        else:
            print("tfidf Training...")
            self.tfidf_model = gensim.models.TfidfModel(corpus=self.corpus, id2word=self.dict)
            print("tfidf Train complete")

    def train_from_doc_collection_with_preprocessor(self, doc_collection: PreprocessMultiFieldDocumentCollection,
                                                    **config):
        pretrain_tfidf_path = None
        pretrain_dictionary_path = None
        if "pretrain_tfidf_path" in config.keys():
            pretrain_tfidf_path = config["pretrain_tfidf_path"]
        if "pretrain_dictionary_path" in config.keys():
            pretrain_dictionary_path = config["pretrain_dictionary_path"]

        if isinstance(doc_collection, PreprocessMultiFieldDocumentCollection):
            self.train_from_document_collection(preprocess_document_collection=doc_collection,
                                                pretrain_tfidf_path=pretrain_tfidf_path,
                                                pretrain_dictionary_path=pretrain_dictionary_path
                                                )

    def save(self, model_dir_path):
        """
        save model to the model_dir_path
        :param model_dir_path: the dir to save the model
        :return:
        """
        super().save(model_dir_path)
        self.__init_sub_model_path__()
        self.dict.save(self.dictionary_path)
        print("build dictionary in %s" % self.dictionary_path)

        gensim.corpora.MmCorpus.serialize(self.corpus_path, self.corpus)
        print("save the corpus to %s" % self.corpus_path)

        self.tfidf_model.save(self.tfidf_model_path)
        print("TFIDF Training finish , save to %s" % self.tfidf_model_path)

        print("TFIDF similarity index init ...")
        self.similarity_tfidf_index = gensim.similarities.Similarity(self.sim_index_path,
                                                                     corpus=self.tfidf_model[self.corpus],
                                                                     num_features=len(self.dict))
        print("TFIDF similarity index complete")

        self.similarity_tfidf_index.save(self.sim_index_path)
        print("TFIDF similarity index finish , save to %s" % self.sim_index_path)

        print("entity collection saving...")
        self.preprocess_doc_collection.save(self.entity_collection_path)
        print(
            "entity collection finish saving , save to %s, %r" % (
                self.entity_collection_path, self.preprocess_doc_collection))

    def get_full_doc_score_vec(self, query):
        """
        score vector is a vector v=[0.5,2.0,3.0], v[0] means that the document 'd'
         whose index is 0 in DocumentCollection
        its score with query is 0.5.
        :param query: a str stands for the query.
        :return: get all document similar score with query as a numpy vector.
        """

        full_entity_score_vec = self.get_cache_score_vector(query)
        if full_entity_score_vec is not None:
            return full_entity_score_vec

        query_words = self.preprocessor.clean(query)
        query_bow = self.dict.doc2bow(query_words)
        query_tfidf_vec = self.tfidf_model[query_bow]
        full_entity_score_vec = self.similarity_tfidf_index[query_tfidf_vec]

        self.cache_entity_score_vector(query, full_entity_score_vec)
        return full_entity_score_vec

    def similarity_by_doc_id(self, doc_id):
        doc_index = self.doc_id2doc_index(doc_id)
        return self.similarity_tfidf_index.similarity_by_id(doc_index)

    def __init_index(self):
        try:
            self.similarity_tfidf_index = gensim.similarities.Similarity.load(self.sim_index_path)
            self.get_full_doc_score_vec("add")
            # todo: change to some keyword must in the dict
        except FileNotFoundError:
            print("the sim index is useless, generate again")
            self.similarity_tfidf_index = gensim.similarities.Similarity(self.sim_index_path,
                                                                         corpus=self.tfidf_model[self.corpus],
                                                                         num_features=len(self.dict))

            self.similarity_tfidf_index.save(self.sim_index_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import traceback
from abc import ABCMeta, abstractmethod
from copy import deepcopy

import gensim
import networkx
from networkx import MultiDiGraph

from sekg.graph.accessor import GraphAccessor
from sekg.graph.creator import NodeBuilder
from sekg.graph.index_accessor import IndexGraphAccessor
from sekg.graph.metadata_accessor import MetadataGraphAccessor
from sekg.util.annotation import catch_exception

DEFAULT_LABEL = "entity"
DEFAULT_PRIMARY_KEY = "_node_id"

"""
this package is used to export a neo4j instance to Memory to build a copy of graph.
You can used it to train graph vector or to query graph 
"""


class NodeInfo(object):
    """
    a basic abstract class for NodeInfo
    """
    __metaclass__ = ABCMeta
    PRIVATE_PROPERTY = {
        "lastrevid",
        "_create_time",
        "_update_time",
        "_modify_version",
        "modified",
        "logo image"
    }

    def __init__(self, node_id, labels, properties):
        self.labels = labels
        self.properties = properties
        self.node_id = node_id

    @abstractmethod
    def get_main_name(self):
        pass

    @abstractmethod
    def get_all_names(self):
        return []

    def get_all_names_str(self):
        return " , ".join(self.get_all_names())

    @abstractmethod
    def get_all_valid_attributes(self):
        valid_attribute_pairs = []

        for property_name in self.properties.keys():
            if self.is_valid_property(property_name):
                value = self.properties[property_name]
                if not value:
                    continue
                valid_attribute_pairs.append((property_name, value))
        return valid_attribute_pairs

    def get_all_valid_attributes_str(self):
        result = []
        valid_attribute_pairs = self.get_all_valid_attributes()
        for (property_name, value) in valid_attribute_pairs:
            result.append(property_name + " : " + value)

        return " , ".join(result)

    @abstractmethod
    def is_valid_property(self, property_name):
        if property_name in NodeInfo.PRIVATE_PROPERTY:
            return False
        return True

    def __repr__(self):
        return "<NodeInfo id=%d labels=%r properties=%r>" % (self.node_id, self.labels, self.properties)


class NodeInfoFactory(object):
    """
    a basic abstract class for NodeInfo
    """
    __metaclass__ = ABCMeta
    """
    a factory to create NodeInfo, can be extend to create Multi type NodeInfo instance by condition
    """

    def __init__(self):
        pass

    @abstractmethod
    def create_node_info(self, node_info_dict):
        """need to implement"""
        return None


class RelationInfo:

    def __init__(self, start_node_info, relation_name, end_node_info):
        self.relation_name = relation_name
        self.end_node_info = end_node_info
        self.start_node_info = start_node_info

    def get_in_relation_str(self):
        return "{start} {r}".format(
            start=self.start_node_info.get_main_name(),
            r=self.relation_name
        )

    def get_out_relation_str(self):
        return "{r} {end}".format(
            r=self.relation_name,
            end=self.end_node_info.get_main_name()
        )

    def get_full_relation_str(self):
        return "{start} {r} {end}".format(start=self.start_node_info.get_main_name(),
                                          r=self.relation_name,
                                          end=self.end_node_info.get_main_name())


class GraphDataReader:
    """
    this class is a graph Data reader, can get the graph exporter not in the raw json,
    but in more clean format, wrap the result to a specific Object NodeInfo or RelationInfo.
    """

    def __init__(self, graph_data, node_info_factory):
        if isinstance(graph_data, GraphData):
            self.graph_data = graph_data
        else:
            self.graph_data = None

        if isinstance(node_info_factory, NodeInfoFactory):
            self.node_info_factory = node_info_factory
        else:
            self.node_info_factory = None

    def get_all_out_relation_infos(self, node_id):
        relation_list = self.graph_data.get_all_out_relations(node_id=node_id)

        return self.create_relation_info_list(relation_list)

    def get_all_in_relation_infos(self, node_id):
        relation_list = self.graph_data.get_all_in_relations(node_id=node_id)

        return self.create_relation_info_list(relation_list)

    @staticmethod
    def create_relation_info(start_node_info, relation_type, end_node_info):
        return RelationInfo(start_node_info=start_node_info,
                            relation_name=relation_type,
                            end_node_info=end_node_info)

    def create_from_relation_info_dict(self, relation_info):
        graph_data = self.graph_data
        start_id, relation_type, end_id = relation_info
        start_node_info_dict = graph_data.get_node_info_dict(start_id)
        start_node_info = self.node_info_factory.create_node_info(start_node_info_dict)

        end_node_info_dict = graph_data.get_node_info_dict(end_id)
        end_node_info = self.node_info_factory.create_node_info(end_node_info_dict)

        return self.create_relation_info(start_node_info=start_node_info, relation_type=relation_type,
                                         end_node_info=end_node_info,
                                         )

    def create_relation_info_list(self, relation_info_dict_list):
        info_list = []
        for r in relation_info_dict_list:
            info_list.append(self.create_from_relation_info_dict(r))
        return info_list

    def get_node_info(self, node_id):
        node_info_dict = self.graph_data.get_node_info_dict(node_id=node_id)
        return self.node_info_factory.create_node_info(node_info_dict)

    def get_all_node_infos(self):
        result = []
        for node_id in self.graph_data.get_node_ids():
            result.append(self.get_node_info(node_id))
        return result


class NodePropertyIndexer:
    """
    a helper class to help the graph data object create the index on some properties.
    And it could find node id with specific property value on the some properties.
    """

    def __init__(self, index_property_name):
        self.index_property_name = index_property_name
        self.property_value_to_ids_map = {}
        self.id_2_property_values_map = {}

    def index_node(self, node_id, node_properties):
        if self.index_property_name not in node_properties:
            return False
        property_value = node_properties[self.index_property_name]

        exist_property_values = self.get_indexed_property_values(node_id)

        if type(property_value) == list or type(property_value) == set:
            new_property_values = set(property_value)
        else:
            new_property_values = {property_value}

        if exist_property_values == new_property_values:
            return False

        delete_property_values = exist_property_values - new_property_values
        add_property_values = new_property_values - exist_property_values

        for t_value in delete_property_values:
            self.remove_index_on_value(node_id=node_id, property_value=t_value)

        for t_value in add_property_values:
            self.add_index_on_value(node_id=node_id, property_value=t_value)

        return True

    def get_indexed_property_values(self, node_id):
        if node_id not in self.id_2_property_values_map:
            self.id_2_property_values_map[node_id] = set([])
        exist_property_values = self.id_2_property_values_map[node_id]
        return exist_property_values

    def find_node_ids_by_value(self, value):
        if value in self.property_value_to_ids_map:
            return self.property_value_to_ids_map[value]
        return set([])

    def remove_index_on_value(self, node_id, property_value):
        if property_value not in self.property_value_to_ids_map:
            return

        id_set = self.property_value_to_ids_map[property_value]
        if node_id in id_set:
            id_set.remove(node_id)

        if node_id not in self.id_2_property_values_map:
            return

        property_values = self.id_2_property_values_map[node_id]
        if property_value in property_values:
            property_values.remove(property_value)

    def add_index_on_value(self, node_id, property_value):
        if property_value not in self.property_value_to_ids_map:
            self.property_value_to_ids_map[property_value] = set([])

        id_set = self.property_value_to_ids_map[property_value]

        id_set.add(node_id)

        if node_id not in self.id_2_property_values_map:
            self.id_2_property_values_map[node_id] = set([])

        property_values = self.id_2_property_values_map[node_id]
        property_values.add(property_value)

    def remove_index_on_node(self, node_id):
        if node_id in self.id_2_property_values_map:
            index_values = self.id_2_property_values_map[node_id]

            all_index_values = list(index_values)
            for value in all_index_values:
                self.remove_index_on_value(node_id=node_id, property_value=value)
            self.id_2_property_values_map.pop(node_id)


class GraphIndexCollection:
    def __init__(self):
        self.property_to_indexer_map = {}

    def create_index_on_property(self, *property_name_list):
        for name in property_name_list:
            if name not in self.property_to_indexer_map:
                self.property_to_indexer_map[name] = NodePropertyIndexer(index_property_name=name)

    def add_node(self, node_id, node_properties):
        for property_name, indexer in self.property_to_indexer_map.items():
            indexer.index_node(node_id, node_properties)

    def remove_node(self, node_id):
        for property_name, indexer in self.property_to_indexer_map.items():
            indexer.remove_index_on_node(node_id)

    def find_ids(self, property_name, property_value):
        if self.is_property_indexed(property_name) == False:
            return set([])
        return self.property_to_indexer_map[property_name].find_node_ids_by_value(property_value)

    def is_property_indexed(self, property_name):
        """
        check if one property indexed
        :param property_name:
        :return:
        """
        if property_name in self.get_index_property():
            return True
        return False

    def get_index_property(self):
        """
        get all indexed property name
        :return:
        """
        return self.property_to_indexer_map.keys()


class GraphData(gensim.utils.SaveLoad):
    # todo: make the information return by graph data are copy by value, not reference
    # todo: in case the change
    """
    the store of a graph data.

    each node is represent as a dict of node info named 'node_json',
    Example Format for 'node_json':

     {
        "id": 1,
        "properties": {"name":"bob","age":1},
        "labels": ["entity","man"]
    }

    """

    DEFAULT_KEY_NODE_ID = "id"  # the key name for the node id, every node must have it.
    DEFAULT_KEY_NODE_PROPERTIES = "properties"  # the key name for the node properties, every node must have it.
    DEFAULT_KEY_NODE_LABELS = "labels"  # the key name for the node labels, every node must have it.

    DEFAULT_KEYS = [DEFAULT_KEY_NODE_ID, DEFAULT_KEY_NODE_PROPERTIES, DEFAULT_KEY_NODE_LABELS]
    UNASSIGNED_NODE_ID = -1  # a node without a id specify, a newly created node, its id is -1

    DEFAULT_KEY_RELATION_START_ID = "startId"
    DEFAULT_KEY_RELATION_TYPE = "relationType"
    DEFAULT_KEY_RELATION_END_ID = "endId"

    DEFAULT_KEY_PROPERTY_QUALIFIED_NAME = "qualified_name"
    DEFAULT_KEY_PROPERTY_ALIAS = "alias"

    def __init__(self):
        # two map for
        self.__init_graph()

    def clear(self):
        self.__init_graph()

    def __init_graph(self):
        self.graph = MultiDiGraph()
        self.max_node_id = 0
        self.label_to_ids_map = {}
        self.index_collection = GraphIndexCollection()

    def create_index_on_property(self, *property_name_list):
        self.index_collection.create_index_on_property(*property_name_list)

    def find_all_shortest_paths(self, startId, endId):
        """
        找到所有的最短路
        :param startId:
        :param endId:
        :return:
        """
        shortest_paths = networkx.all_shortest_paths(self.graph, startId, endId)
        return shortest_paths

    def find_shortest_path(self, startId, endId):
        """
        找到一个最短路
        :param startId:
        :param endId:
        :return:
        """
        shortest_paths = networkx.shortest_path(self.graph, startId, endId)
        return shortest_paths

    def set_nodes(self, nodes):
        for n in nodes:
            self.add_node(node_id=n[self.DEFAULT_KEY_NODE_ID],
                          node_properties=n[self.DEFAULT_KEY_NODE_PROPERTIES],
                          node_labels=n[self.DEFAULT_KEY_NODE_LABELS])

    def add_labels(self, *labels):
        """
        add a list of label to the graph
        :param labels:
        :return:
        """

        for label in labels:
            if not label:
                return
            if label not in self.label_to_ids_map.keys():
                self.label_to_ids_map[label] = set([])

    def add_label_by_node_id(self, node_id, label):
        """
        add a label to a node
        :param node_id: the node id which the label need to add
        :param label: the label that need to added
        :return: True, add successful.False, add fail.
        """
        if not label:
            return False
        node_json = self.get_node_info_dict(node_id)
        if not node_json:
            return False
        node_json[GraphData.DEFAULT_KEY_NODE_LABELS].add(label)
        self.label_to_ids_map[label].add(node_id)
        return True

    def get_node_ids_by_label(self, label):
        if label not in self.label_to_ids_map.keys():
            return set([])
        return self.label_to_ids_map[label]

    def add_label_by_label(self, label, new_label):
        """
        add a label to node in graph, the node must has the specific label
        :param new_label: the new_label add to node
        :param label: the node must has the label
        :return:
        """

        for node_id in self.get_node_ids_by_label(label):
            self.add_label_by_node_id(node_id, new_label)

    def add_label_to_all(self, label):
        """
        add a label to node in graph
        :param label:
        :return:
        """
        if not label:
            return
        self.add_labels(label)
        for node_id in self.get_node_ids():
            self.add_label_by_node_id(node_id, label)

    def add_node(self, node_labels, node_properties, node_id=UNASSIGNED_NODE_ID, primary_property_name=""):
        """
        add a node json to the graph
        :param node_id: the node_id to identify the node, if not given, it will be add as new node and give a node id
        :param node_properties: a dict of node properties, key-value pair
        :param node_labels: a set of node labels
        :param primary_property_name:make sure the node_json["properties"][primary_property_name] is unique in GraphData.
         if no passing, the node json will be add to graph without check. otherwise, only the node json
        with unique property value ( property value is got by primary_property_name ) will be added to the GraphData.
                :return:-1, means that adding node json fail. otherwise, return the id of the newly added node
        """

        if primary_property_name:
            if primary_property_name not in node_properties:
                print("node json must have a primary_property_name ( %r ) in properties " % primary_property_name)
                return self.UNASSIGNED_NODE_ID

            node_json = self.find_one_node_by_property(property_name=primary_property_name,
                                                       property_value=node_properties[
                                                           primary_property_name])
            if node_json:
                return node_json[self.DEFAULT_KEY_NODE_ID]

        if node_id == self.UNASSIGNED_NODE_ID:
            node_id = self.max_node_id + 1
            self.max_node_id = self.max_node_id + 1

        new_node_json = {
            self.DEFAULT_KEY_NODE_ID: node_id,
            self.DEFAULT_KEY_NODE_PROPERTIES: node_properties,
            self.DEFAULT_KEY_NODE_LABELS: set(node_labels)
        }

        self.graph.add_node(node_id, **new_node_json)

        if self.max_node_id < node_id:
            self.max_node_id = node_id

        self.add_labels(*new_node_json[self.DEFAULT_KEY_NODE_LABELS])
        for label in new_node_json[self.DEFAULT_KEY_NODE_LABELS]:
            self.label_to_ids_map[label].add(node_id)
        self.index_collection.add_node(node_id=node_id,
                                       node_properties=new_node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES])
        return node_id

    def remove_node(self, node_id):
        if node_id not in self.graph.nodes:
            return None
        # print(type(self.graph.nodes))
        node_json = self.graph.nodes[node_id]
        in_relations = set(self.graph.in_edges(node_id, keys=True))
        out_relations = set(self.graph.out_edges(node_id, keys=True))
        self.graph.remove_node(node_id)

        for label in node_json[self.DEFAULT_KEY_NODE_LABELS]:
            self.label_to_ids_map[label].remove(node_id)

        self.index_collection.remove_node(node_id)

        return node_json, out_relations, in_relations

    def merge_two_nodes_by_id(self, main_node_id, merged_node_id,
                              main_alias_property=DEFAULT_KEY_PROPERTY_QUALIFIED_NAME,
                              merged_alias_property=DEFAULT_KEY_PROPERTY_QUALIFIED_NAME, merge_properties=True):
        if main_node_id not in self.graph.nodes or merged_node_id not in self.graph.nodes:
            return None

        main_node_json, main_out_relations, main_in_relations = self.remove_node(main_node_id)
        merged_node_json, merged_out_relations, merged_in_relations = self.remove_node(merged_node_id)
        new_node_labels = main_node_json[self.DEFAULT_KEY_NODE_LABELS] | merged_node_json[self.DEFAULT_KEY_NODE_LABELS]
        new_node_properties = main_node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
        alias_set = set()

        alias = new_node_properties.get(main_alias_property, set())
        if len(alias):
            alias_set.update(alias)
            # alias_set.add(alias)

        alias = merged_node_json[self.DEFAULT_KEY_NODE_PROPERTIES].get(merged_alias_property, set())
        if len(alias):
            alias_set.update(alias)
            # alias_set.add(alias)

        if len(alias_set) > 0:
            if self.DEFAULT_KEY_PROPERTY_ALIAS not in new_node_properties:
                new_node_properties[self.DEFAULT_KEY_PROPERTY_ALIAS] = set()
            new_node_properties[self.DEFAULT_KEY_PROPERTY_ALIAS].update(alias_set)

        if merge_properties:
            for k, v in merged_node_json[self.DEFAULT_KEY_NODE_PROPERTIES].items():
                if k not in new_node_properties:
                    new_node_properties[k] = v
        new_node_id = self.add_node(new_node_labels, new_node_properties)

        new_relations = set()
        for startId, endId, relationType in (
                main_out_relations | main_in_relations | merged_out_relations | merged_in_relations):
            if startId == main_node_id or startId == merged_node_id:
                startId = new_node_id
            if endId == main_node_id or endId == merged_node_id:
                endId = new_node_id
            new_relations.add((startId, relationType, endId))
        new_relations = filter(lambda r: r[0] != r[2], new_relations)
        for startID, relationType, endId in new_relations:
            self.add_relation(startID, relationType, endId)

        return new_node_id

    def merge_node(self, node_labels, node_properties, primary_property_name):
        """
        merge a node json to the graph, that is if we can't not find the node with primary_property_value match the given node.
        we will add a new node, if we found, we will add copy all properties given to the exist node, copy all labels to the exist node.
        properties will be updated by this merge. That is, if the node to be merged has the same attributes as the existing node, the attributes of the new node are used.

        :param node_properties: a dict of node properties, key-value pair
        :param node_labels: a set of node labels
        :param primary_property_name: The name of the property to check, the merged node and the new node are the same on this property.
        :return:-1, means that adding node json fail. otherwise, return the id of the newly added(merged) node.If it already exists, the id of this merged node will not change.
        """

        if not primary_property_name:
            print("primary_property_name must given on merge")
            return GraphData.UNASSIGNED_NODE_ID

        if primary_property_name not in node_properties:
            print("node json must have a primary_property_name ( %r ) in properties " % primary_property_name)
            return self.UNASSIGNED_NODE_ID

        node_json = self.find_one_node_by_property(property_name=primary_property_name,
                                                   property_value=node_properties[
                                                       primary_property_name])

        if not node_json:
            return self.add_node(node_labels=node_labels, node_properties=node_properties,
                                 node_id=GraphData.UNASSIGNED_NODE_ID)

        merge_node_id = node_json[self.DEFAULT_KEY_NODE_ID]
        merge_properties = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
        for k, v in node_properties.items():
            merge_properties[k] = v

        merge_labels = set(node_json[self.DEFAULT_KEY_NODE_LABELS])
        for label in node_labels:
            merge_labels.add(label)

        return self.add_node(node_labels=merge_labels, node_properties=merge_properties, node_id=merge_node_id)

    def add_node_with_multi_primary_property(self, node_labels, node_properties, node_id=UNASSIGNED_NODE_ID,
                                             primary_property_names=None):
        """
        add a node json to the graph
        :param node_id: the node_id to identify the node, if not given, it will be add as new node and give a node id
        :param node_properties: a dict of node properties, key-value pair
        :param node_labels: a set of node labels
        :param primary_property_names:a list of primary properties. make sure the node_json["properties"][primary_property_name] is unique in GraphData.
         if no passing, the node json will be add to graph without check. otherwise, only the node json
        with unique property value ( property value is got by primary_property_name ) will be added to the GraphData.
                :return:-1, means that adding node json fail. otherwise, return the id of the newly added node
        """

        if primary_property_names is None:
            primary_property_names = []

        match_properties = {}

        for primary_property_name in primary_property_names:
            if primary_property_name not in node_properties:
                print("node json must have a primary_property_name ( %r ) in properties " % primary_property_name)
                return self.UNASSIGNED_NODE_ID
            match_properties[primary_property_name] = node_properties[primary_property_name]

        node_json = self.find_one_node_by_properties(**match_properties)
        if node_json:
            return node_json[self.DEFAULT_KEY_NODE_ID]

        if node_id == self.UNASSIGNED_NODE_ID:
            node_id = self.max_node_id + 1

        new_node_json = {
            self.DEFAULT_KEY_NODE_ID: node_id,
            self.DEFAULT_KEY_NODE_PROPERTIES: node_properties,
            self.DEFAULT_KEY_NODE_LABELS: set(node_labels)
        }

        self.graph.add_node(node_id, **new_node_json)
        if self.max_node_id < node_id:
            self.max_node_id = node_id

        self.add_labels(*new_node_json[self.DEFAULT_KEY_NODE_LABELS])
        for label in new_node_json[self.DEFAULT_KEY_NODE_LABELS]:
            self.label_to_ids_map[label].add(node_id)
        self.index_collection.add_node(node_id=node_id,
                                       node_properties=new_node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES])

        return node_id

    def merge_node_with_multi_primary_property(self, node_labels, node_properties, primary_property_names=None):
        """
        merge a node json to the graph, that is if we can't not find the node with primary_property_value match the given node.
        we will add a new node, if we found, we will add copy all properties given to the exist node, copy all labels to the exist node.
        properties will be updated by this merge. That is, if the node to be merged has the same attributes as the existing node, the attributes of the new node are used.

        :param node_properties: a dict of node properties, key-value pair
        :param node_labels: a set of node labels
        :param primary_property_name: The list of name of the property to check, the merged node and the new node are the same on this property.
        :return:-1, means that adding node json fail. otherwise, return the id of the newly added(merged) node.If it already exists, the id of this merged node will not change.
        """

        if not primary_property_names:
            print("primary_property_names must given on merge")
            return GraphData.UNASSIGNED_NODE_ID
        match_properties = {}

        for primary_property_name in primary_property_names:
            if primary_property_name not in node_properties:
                print("node json must have a primary_property_name ( %r ) in properties " % primary_property_name)
                return self.UNASSIGNED_NODE_ID
            match_properties[primary_property_name] = node_properties[primary_property_name]

        node_json = self.find_one_node_by_properties(**match_properties)
        if not node_json:
            return self.add_node(node_labels=node_labels, node_properties=node_properties,
                                 node_id=GraphData.UNASSIGNED_NODE_ID)

        merge_node_id = node_json[self.DEFAULT_KEY_NODE_ID]
        merge_properties = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
        for k, v in node_properties.items():
            merge_properties[k] = v

        merge_labels = set(node_json[self.DEFAULT_KEY_NODE_LABELS])
        for label in node_labels:
            merge_labels.add(label)

        return self.add_node(node_labels=merge_labels, node_properties=merge_properties, node_id=merge_node_id)

    def refresh_indexer(self):
        """
        refresh the index on all properties.
        :return:
        """
        index_properties = self.index_collection.get_index_property()
        index_properties = list(index_properties)
        del self.index_collection
        self.index_collection = GraphIndexCollection()

        self.create_index_on_property(*index_properties)
        for node_id, node_json in self.graph.nodes(data=True):
            if node_json is None:
                continue
            node_properties_json = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
            self.index_collection.add_node(node_id, node_properties_json)

    def find_one_node_by_property(self, property_name, property_value):
        if self.index_collection.is_property_indexed(property_name):
            candidate_node_ids = list(self.index_collection.find_ids(property_name, property_value=property_value))
            if len(candidate_node_ids) == 0:
                return None
            return self.get_node_info_dict(candidate_node_ids[0])

        for node_id, node_json in self.graph.nodes(data=True):
            node_properties_json = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
            if property_name in node_properties_json.keys() and node_properties_json[property_name] == property_value:
                return node_json
        return None

    def find_nodes_by_ids(self, *ids):
        result = []
        for node_id in ids:
            node_json = self.get_node_info_dict(node_id)
            if node_json:
                result.append(node_json)
        return result

    def find_nodes_by_property(self, property_name, property_value):
        if self.index_collection.is_property_indexed(property_name):
            candidate_node_ids = list(self.index_collection.find_ids(property_name, property_value=property_value))

            return self.find_nodes_by_ids(*candidate_node_ids)

        nodes = []
        for node_id, node_json in self.graph.nodes(data=True):
            node_properties_json = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
            if property_name in node_properties_json.keys() and node_properties_json[property_name] == property_value:
                nodes.append(node_json)
        return nodes

    def find_one_node_by_property_value_starts_with(self, property_name, property_value_starter):
        """
        find a node which its property value is string and the string is startswith a given string
        :param property_name:
        :param property_value_starter:
        :return:
        """
        for node_id, node_json in self.graph.nodes(data=True):
            node_properties_json = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
            if property_name not in node_properties_json.keys():
                continue

            property_value = node_properties_json[property_name]
            if type(property_value) != str:
                continue
            if property_value.startswith(property_value_starter):
                return node_json
        return None

    def find_nodes_by_property_value_starts_with(self, property_name, property_value_starter):
        """
        find all nodes which its property value is string and the string is startswith a given string
        :param property_name:
        :param property_value_starter:
        :return:
        """
        nodes = []
        for node_id, node_json in self.graph.nodes(data=True):
            node_properties_json = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]
            if property_name not in node_properties_json.keys():
                continue

            property_value = node_properties_json[property_name]
            if type(property_value) != str:
                continue
            if property_value.startswith(property_value_starter):
                nodes.append(node_json)
        return nodes

    def __find_node_ids_by_index_properties(self, **index_properties):
        result_ids = self.get_node_ids()

        for property_name, property_value in index_properties.items():
            result_ids = result_ids.intersection(
                self.index_collection.find_ids(property_name=property_name,
                                               property_value=property_value))

        return result_ids

    def find_one_node_by_properties(self, **properties):
        indexed_properties = {}
        unindexed_properties = {}
        for property_name, property_value in properties.items():
            if self.index_collection.is_property_indexed(property_name=property_name):
                indexed_properties[property_name] = property_value
            else:
                unindexed_properties[property_name] = property_value

        candidate_node_ids = self.__find_node_ids_by_index_properties(**indexed_properties)

        if len(candidate_node_ids) == 0:
            return None

        if len(unindexed_properties) == 0:
            return self.get_node_info_dict(list(candidate_node_ids)[0])

        for node_id in candidate_node_ids:
            node_json = self.get_node_info_dict(node_id=node_id)
            node_properties_json = node_json[self.DEFAULT_KEY_NODE_PROPERTIES]

            is_match = True
            for property_name, property_value in unindexed_properties.items():
                if property_name not in node_properties_json.keys() or node_properties_json[
                    property_name
                ] != property_value:
                    is_match = False
                    break
            if is_match:
                return node_json

        return None

    def set_relations(self, relations):
        for t in relations:
            self.add_relation(startId=t[self.DEFAULT_KEY_RELATION_START_ID],
                              relationType=t[self.DEFAULT_KEY_RELATION_TYPE],
                              endId=t[self.DEFAULT_KEY_RELATION_END_ID])

    def add_relation(self, startId, relationType, endId):
        """
        add a new relation to graphData, if exist, not add.
        :param startId:
        :param relationType:
        :param endId:
        :return:False, the relation is already exist adding fail, True, add the relation successsful
        """
        # if startId == GraphData.UNASSIGNED_NODE_ID:
        #     return False
        # if endId == GraphData.UNASSIGNED_NODE_ID:
        #     return False

        if startId not in self.graph.nodes or endId not in self.graph.nodes:
            return False

        if self.exist_relation(startId=startId, relationType=relationType, endId=endId):
            return False

        self.graph.add_edge(startId, endId, relationType)
        return True

    def add_relation_with_property(self, startId, relationType, endId, **kwargs):
        if startId not in self.graph.nodes or endId not in self.graph.nodes:
            return False

        if self.exist_relation(startId=startId, relationType=relationType, endId=endId):
            return False

        self.graph.add_edge(startId, endId, relationType, **kwargs)
        return True

    def remove_relation(self, startId, relationType, endId):
        if not self.exist_relation(startId=startId, relationType=relationType, endId=endId):
            return False

        self.graph.remove_edge(startId, endId, relationType)
        return True

    def exist_relation(self, startId, relationType, endId):
        return self.graph.has_edge(startId, endId, relationType)

    def get_relations(self, start_id=None, relation_type=None, end_id=None):
        candidates = None
        if start_id is not None:
            candidates = self.get_all_out_relations(start_id)
        if end_id is not None:
            tmp = self.get_all_in_relations(end_id)
            if candidates is not None:
                candidates &= tmp
            else:
                candidates = tmp
        candidates = self.get_relation_pairs_with_type() if candidates is None else candidates

        if relation_type is not None:
            candidates = set(filter(lambda r: r[1] == relation_type, candidates))
        return candidates



    def get_edge_extra_info(self, start_id, end_id, relation_name, extra_key):
        relation_dict = self.graph.get_edge_data(start_id, end_id)
        if relation_name in relation_dict:
            if extra_key in relation_dict[relation_name]:
                return relation_dict[relation_name][extra_key]
        return ""

    def save_as_json(self, path):
        pass
        # temp = {
        #     "out_relation_map": self.out_relation_map,
        #     "in_relation_map": self.in_relation_map,
        #     "id_to_nodes_map": self.id_to_node_map
        # }
        # json.dump(temp, path)

    def get_node_num(self):
        return len(self.graph.nodes)

    def get_relation_num(self):
        return len(self.graph.edges)

    def get_node_ids(self):
        return set(self.graph.nodes)

    def get_relation_pairs(self):
        # todo:cache the result?
        """
        get the relation list in [(startId,endId)] format
        :return:
        """
        pairs = set(self.graph.edges(keys=False))

        return pairs

    def get_relation_pairs_with_type(self):
        """
        get the relation list in [(startId,endId)] format
        :return:
        """
        pairs = {(r[0], r[2], r[1]) for r in self.graph.edges(keys=True)}
        return pairs

    def get_all_out_relations(self, node_id):
        if node_id not in self.graph.nodes:
            return set()
        return {(r[0], r[2], r[1]) for r in self.graph.out_edges(node_id, keys=True)}

    def get_all_in_relations(self, node_id):
        if node_id not in self.graph.nodes:
            return set()
        return {(r[0], r[2], r[1]) for r in self.graph.in_edges(node_id, keys=True)}

    def update_node_index(self, node_id):

        node_info = self.get_node_info_dict(node_id=node_id)
        node_properties = node_info[self.DEFAULT_KEY_NODE_PROPERTIES]
        self.index_collection.add_node(node_id=node_id
                                       , node_properties=node_properties)

    def get_node_info_dict(self, node_id):
        """
        get the node info dict,
        :param node_id: the node id
        :return:
        """
        return self.graph.nodes.get(node_id, None)

    def get_all_labels(self):
        """
        get all labels as set for current node.
        :return: a set of labels.
        """
        return self.label_to_ids_map.keys()

    def print_label_count(self):
        print("Label Num=%d" % len(self.label_to_ids_map.keys()))
        for k, v in self.label_to_ids_map.items():
            print("<Label:%r Num:%d>" % (k, len(v)))

    def print_graph_info(self):
        print("----- Graph Info ------")
        print(self)
        self.print_label_count()
        self.print_relation_info()
        print("-----------------------")

    def print_relation_info(self):
        relation_type_to_relation = {}
        for r in self.get_relation_pairs_with_type():
            if r[1] not in relation_type_to_relation:
                relation_type_to_relation[r[1]] = set()
            relation_type_to_relation[r[1]].add(r)
        print("Relation Num=%d" % len(relation_type_to_relation.keys()))
        for k, v in relation_type_to_relation.items():
            print("<Relation:%r Num:%d>" % (k, len(v)))

    def __repr__(self):
        return "<GraphData nodeNum=%d relNum=%d maxNodeId=%d>" % (
            self.get_node_num(), self.get_relation_num(), self.max_node_id)

    def subgraph(self, node_ids):
        """
        get a sub graph of graph data which keep only given nodes and relations between nodes
        :param node_ids: the kept node ids in graph
        :return: a graph that keep all things.
        """
        graph_data = deepcopy(self)

        remove_nodes = set(self.get_node_ids()) - node_ids
        for node_id in remove_nodes:
            graph_data.remove_node(node_id)

        return graph_data


class DataExporterAccessor(GraphAccessor):
    def get_all_nodes_not_batch(self, node_label):
        try:
            query = 'Match (n:`{node_label}`) return n'.format(node_label=node_label)

            cursor = self.graph.run(query)
            nodes = []
            for record in cursor:
                nodes.append(record["n"])
            return nodes
        except Exception:
            return []

    def get_all_nodes(self, node_label, step=100000):
        metadata_accessor = MetadataGraphAccessor(self)
        max_node_id = metadata_accessor.get_max_id_for_node()

        nodes = []

        for start_id in range(0, max_node_id, step):
            end_id = min(max_node_id, start_id + step)
            nodes.extend(self.get_all_nodes_in_scope(node_label, start_id=start_id, end_id=end_id))
            print("get nodes step %d-%d" % (start_id, end_id))

        return nodes

    def get_all_nodes_in_scope(self, node_label, start_id, end_id):
        try:
            query = 'Match (n:`{node_label}`) where ID(n)>{start_id} and ID(n)<={end_id} return n'.format(
                node_label=node_label, start_id=start_id, end_id=end_id)

            cursor = self.graph.run(query)
            nodes = []
            for record in cursor:
                nodes.append(record["n"])

            return nodes
        except Exception:
            return []

    def get_all_relation(self, node_label, step=200000):
        metadata_accessor = MetadataGraphAccessor(self)
        max_relation_id = metadata_accessor.get_max_id_for_relation()

        relations = []

        for start_id in range(0, max_relation_id, step):
            end_id = min(max_relation_id, start_id + step)
            relations.extend(self.get_all_relation_in_scope(node_label, start_id=start_id, end_id=end_id))
            print("get relation step %d-%d" % (start_id, end_id))
        return relations

    def get_all_relation_in_scope(self, node_label, start_id, end_id):
        try:
            query = 'Match (n:`{node_label}`)-[r]->(m:`{node_label}`) where ID(r)>{start_id} and ID(r)<={end_id} return ID(n) as startId,ID(m) as endId, type(r) as relationType'.format(
                node_label=node_label, start_id=start_id, end_id=end_id)

            cursor = self.graph.run(query)
            data = cursor.data()
            return data
        except Exception:
            return []

    def get_all_relation_not_batch(self, node_label):
        try:
            query = 'Match (n:`{node_label}`)-[r]->(m:`{node_label}`) return ID(n) as startId,ID(m) as endId, type(r) as relationType'.format(
                node_label=node_label)

            cursor = self.graph.run(query)
            data = cursor.data()
            return data
        except Exception:
            return []


class Neo4jImporter:
    """
    The class is used to import one GraphData obj to a Neo4j database.
    """

    def __init__(self, graph_client: GraphAccessor):
        self.graph_accessor = graph_client

    def import_all_graph_data(self, graph_data: GraphData, clear=True):
        """
        import all data in one GraphData into neo4j and create index on node
        :param graph_data:
        :param clear: clear the graph content, default is not clear the graph contain
        :return:
        """
        index_accessor = IndexGraphAccessor(self.graph_accessor)
        index_accessor.create_index(label=DEFAULT_LABEL, property_name=DEFAULT_PRIMARY_KEY)

        if clear:
            self.graph_accessor.delete_all_relations()
            self.graph_accessor.delete_all_nodes()

        # todo: this is slow, need to speed up, maybe not commit on every step
        all_node_ids = graph_data.get_node_ids()
        for node_id in all_node_ids:
            ## todo: fix this by not using 'properties','labels'
            node_info_dict = graph_data.get_node_info_dict(node_id)
            properties = node_info_dict['properties']
            labels = node_info_dict['labels']
            self.import_one_entity(node_id, properties, labels)

        print("all entity imported")
        relations = graph_data.get_relations()
        for r in relations:
            start_node_id, r_name, end_node_id = r
            start_node = self.graph_accessor.find_node(primary_label=DEFAULT_LABEL,
                                                       primary_property=DEFAULT_PRIMARY_KEY,
                                                       primary_property_value=start_node_id)
            end_node = self.graph_accessor.find_node(primary_label=DEFAULT_LABEL, primary_property=DEFAULT_PRIMARY_KEY,
                                                     primary_property_value=end_node_id)

            if start_node is not None and end_node is not None:
                try:
                    self.graph_accessor.create_relation_without_duplicate(start_node, r_name, end_node)
                except Exception as e:
                    traceback.print_exc()
            else:
                print("fail create relation because start node or end node is none.")
        print("all relation imported")

        print("all graph data import finish")

    @catch_exception
    def import_one_entity(self, node_id, property_dict, labels):
        builder = NodeBuilder()
        builder.add_label(DEFAULT_LABEL).add_property(**property_dict). \
            add_one_property(property_name=DEFAULT_PRIMARY_KEY, property_value=node_id).add_labels(*labels)
        node = builder.build()

        node = self.graph_accessor.create_or_update_node(node, primary_label=DEFAULT_LABEL,
                                                         primary_property=DEFAULT_PRIMARY_KEY)


class GraphDataExporter:
    """
    export specific data
    """

    ## todo: this class should be a opponent class Neo4jImporter, it export Neo4j data as a GraphData obj
    ## todo: take care of the "node_id" problem,
    def __init__(self):
        pass

    def export_all_graph_data(self, graph, node_label):
        accessor = DataExporterAccessor(graph=graph)
        nodes = accessor.get_all_nodes(node_label=node_label)
        graph_data = GraphData()

        for node in nodes:
            labels = [label for label in node.labels]
            graph_data.add_node(node_id=node.identity, node_labels=labels, node_properties=dict(node))

        print("load entity complete, num=%d" % len(nodes))
        relations = accessor.get_all_relation(node_label=node_label)
        print("load relation complete,num=%d" % len(relations))
        graph_data.set_relations(relations=relations)

        return graph_data



from unittest import TestCase

import numpy as np

from sekg.util.vector_util import VectorUtil


class TestVectorUtil(TestCase):
    def test_get_weight_mean_vec(self):
        vectors = [
            np.array([10, 10, 10, 10]),
            np.array([10, 90, 90, 90]),
        ]
        result = VectorUtil.get_weight_mean_vec(vectors, weight_list=[1, 1])

        for right, t in zip(np.array([10, 50, 50, 50]), result):
            self.assertEqual(right, t)
        result = VectorUtil.get_weight_mean_vec(vectors, weight_list=[1, 9])
        for right, t in zip(np.array([10, 82, 82, 82]), result):
            self.assertEqual(right, t)

    def test_compute_idf_weight_dict(self):
        number_dict = {
            1: 10,
            2: 1,
            3: 20
        }

        result = VectorUtil.compute_idf_weight_dict(100, number_dict)

        result_number_dict = {
            1: 2.302585092994046,
            2: 4.605170185988092,
            3: 1.6094379124341003
        }
        for key, v in result_number_dict.items():
            self.assertEqual(v, result[key])



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

from sekg.meta import __author__, __email__, __license__, __package__, __version__

packages = find_packages(exclude=("docs", "test"))

setup(
    name=__package__,
    version=__version__,
    keywords=("pip", "kg", "se"),
    description="knowledge graph util for software engineering",
    long_description="knowledge graph util for software engineering",
    license=__license__,

    url="https://github.com/FudanSELab/sekg",
    author=__author__,
    author_email=__email__,

    packages=packages,
    package_data={
        # If any package contains *.json files, include them:
        '': ['*.json'],
    },
    platforms="any",
    install_requires=[
        "py2neo>=4.1.3",
        "sqlalchemy",
        "pymysql",
        "beautifulsoup4",
        "gensim>=3.8",
        "networkx==2.2",
        "numpy>=1.11.2",
        "spacy>=2.1.3",
        "aiohttp",
        "async_timeout",
        "nltk",
        "kgtools",
        "textdistance[extras]"
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from sekg.term.abbr import AbbrExtractor

if __name__ == "__main__":
    extractor = AbbrExtractor()
    terms, pairs = extractor.extract("For example, feed forward neural networks are comprised of dense layers, while recurrent neural networks can include Graves LSTM (long short-term memory) layers.")
    print(terms, pairs)


#!/usr/bin/python
# -*- coding: UTF-8 -*-
import pickle
from pathlib import Path

import numpy as np
from gensim.corpora import Dictionary, MmCorpus
from gensim.models import TfidfModel
from gensim.models.keyedvectors import Word2VecKeyedVectors

from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.models.base import DocumentSimModel
from sekg.ir.models.util.load import LoadUtil
from sekg.util.vector_util import VectorUtil


class FilterTFIDFNode2VectorModel(DocumentSimModel):
    """
    filter some node by a pretrained Search Model(TFIDFSearchModel)"""
    __corpus_name__ = 'corpus.mm'
    __n2v_model_name__ = "node2vec.model"
    __dictionary_name__ = 'dictionary.dict'
    __entity_collection_name__ = 'entity.collection'
    __kg_name_searcher_name__ = 'kg.namesearcher'
    __graph_data_name__ = "kg.graph"
    __tfidf_model_name__ = "tfidf.model"

    __model_config__ = "model.config"

    __doc_sim_model_dir_name__ = "doc_sim_model"
    DEFAULT_EMBEDDING_SIZE = 100
    MIN_NODE_SIM_SCORE = 0.5
    TOP_N = 1000
    MAX_CANDIDATE_NODE_NUMBER_FOR_ONE_WORD = 10000

    def __init__(self, name, model_dir_path, **config):

        """
        init the lsi model with
        :param model_dir_path:
        """
        super().__init__(name, model_dir_path, **config)
        self.node2vec_model = None

        self.corpus = None
        self.dict = None
        self.tfidf_model = None
        self.similarity_index = None

        self.kg_name_searcher = None
        self.graph_data = None

        self.__init_sub_model_path__()

        self.preprocessor = None
        self.preprocess_doc_collection = None
        self.NP_VECTOR_NOT_EXIST = None
        self.embedding_size = 0
        self.__init_embedding_size(self.DEFAULT_EMBEDDING_SIZE)

        self.doc_sim_model_path = None
        self.doc_sim_model_class = None
        self.doc_sim_model = None

    def load_extra_config(self):
        """
        load some config from pickle file
        :return:
        """
        with open(self.extra_config_path, 'rb') as aq:
            extra_config = pickle.loads(aq.read())

        self.doc_sim_model_class = extra_config["doc_sim_model_class"]
        self.embedding_size = extra_config["embedding_size"]

    def save_extra_config(self):
        """
        save some config into pickle file
        :return:
        """
        extra_config = {}
        extra_config["doc_sim_model_class"] = self.doc_sim_model_class
        extra_config["embedding_size"] = self.embedding_size

        with open(self.extra_config_path, 'wb') as out:
            out.write(pickle.dumps(extra_config))

    def __init_sub_model_path__(self):
        if self.model_dir_path is None:
            return

        model_dir_path_obj = Path(self.model_dir_path)

        # init the paths of models
        self.corpus_path = str(model_dir_path_obj / self.__corpus_name__)

        self.dictionary_path = str(model_dir_path_obj / self.__dictionary_name__)
        self.n2v_model_path = str(model_dir_path_obj / self.__n2v_model_name__)
        self.graph_data_path = str(model_dir_path_obj / self.__graph_data_name__)
        self.entity_collection_path = str(model_dir_path_obj / self.__entity_collection_name__)
        self.kg_name_searcher_path = str(model_dir_path_obj / self.__kg_name_searcher_name__)

        self.tfidf_model_path = str(model_dir_path_obj / self.__tfidf_model_name__)
        self.doc_sim_model_path = str(model_dir_path_obj / self.__doc_sim_model_dir_name__)

        Path(self.doc_sim_model_path).mkdir(exist_ok=True, parents=True)

        self.extra_config_path = str(model_dir_path_obj / self.__model_config__)

    def __init_document_collection(self, preprocess_doc_collection):
        self.preprocess_doc_collection = preprocess_doc_collection
        self.preprocessor = preprocess_doc_collection.get_preprocessor()
        self.field_set = preprocess_doc_collection.get_field_set()

    def __init_embedding_size(self, embedding_size):
        self.embedding_size = embedding_size
        self.NP_VECTOR_NOT_EXIST = np.zeros(embedding_size)
        self.NP_VECTOR_NOT_EXIST[0] = 1e-07

    def init_model_as_submodel(self):
        """
        init the model
        :return:
        """
        print("loading doc_collection")
        self.__init_sub_model_path__()
        preprocess_doc_collection: PreprocessMultiFieldDocumentCollection = PreprocessMultiFieldDocumentCollection.load(
            self.entity_collection_path)
        self.preprocessor = preprocess_doc_collection.get_preprocessor()

        print("loading the tfidf model models")
        self.tfidf_model = TfidfModel.load(self.tfidf_model_path)

        print("loading the Node2vec models")
        self.node2vec_model = Word2VecKeyedVectors.load(self.n2v_model_path)

        print("loading the KGNameSearcher models")

        self.kg_name_searcher = KGNameSearcher.load(self.kg_name_searcher_path)

        self.dict = Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)

        self.load_extra_config()
        self.doc_sim_model = self.doc_sim_model_class.load_as_submodel(self.doc_sim_model_path)

        print("loading %r model from %s" % (self.doc_sim_model_class.model_type(), self.doc_sim_model_path))

    def init_model(self):
        """
        init the model
        :return:
        """
        # todo: fix this load as submodel problem, change to load a clean model,
        #   then only load and init necessary data to complete the task, reduce the time and memory need
        print("loading doc_collection")
        self.__init_sub_model_path__()
        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.__init_document_collection(preprocess_doc_collection)

        print("loading the tfidf model models")
        self.tfidf_model = TfidfModel.load(self.tfidf_model_path)

        print("loading the Node2vec models")
        self.node2vec_model = Word2VecKeyedVectors.load(self.n2v_model_path)

        print("loading the KGNameSearcher models")

        self.kg_name_searcher = KGNameSearcher.load(self.kg_name_searcher_path)
        self.corpus = MmCorpus(self.corpus_path)
        self.dict = Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)

        self.load_extra_config()
        self.doc_sim_model = self.doc_sim_model_class.load_as_submodel(self.doc_sim_model_path)

        print("loading %r model from %s" % (self.doc_sim_model_class.model_type(), self.doc_sim_model_path))

    def get_full_doc_score_vec(self, query):
        full_entity_score_vec = self.get_cache_score_vector(query)
        if full_entity_score_vec is not None:
            return full_entity_score_vec

        query_vec = self.compute_query_graph_vec(query)

        full_entity_score_vec = self.node2vec_model.similar_by_vector(query_vec, topn=None)
        full_entity_score_vec = (full_entity_score_vec + 1) / 2
        self.cache_entity_score_vector(query, full_entity_score_vec)
        return full_entity_score_vec

    def compute_query_graph_vec(self, query):
        words = self.preprocessor.extract_words_for_query(query)
        word_tfidf_weight_dict = {}

        query_bow, query_oov_dict = self.dict.doc2bow(words, return_missing=True)
        query_tfidf_vec = self.tfidf_model[query_bow]

        for word_id, tfidf_value in query_tfidf_vec:
            word = self.dict[word_id]
            word_tfidf_weight_dict[word] = tfidf_value

        tfidf_full_entity_score_vec = self.doc_sim_model.get_full_doc_score_vec(query)
        sort_vec = np.argsort(-tfidf_full_entity_score_vec)
        sort_vec_index = sort_vec[0:self.TOP_N]
        tfidf_full_entity_score_vec = tfidf_full_entity_score_vec[sort_vec_index]
        vectorized_simvec_index2entity_id = np.vectorize(self.doc_index2doc_id)
        index_array = vectorized_simvec_index2entity_id(sort_vec_index)
        valid_id_scores_list = np.array((index_array, tfidf_full_entity_score_vec)).T
        # index_scores = [(index, score) for index, score in enumerate(tfidf_full_entity_score_vec) if
        #                 score > self.MIN_NODE_SIM_SCORE]
        valid_id_set = set([entity_id for (entity_id, score) in valid_id_scores_list])
        entity_weight_dict = {}
        for word in word_tfidf_weight_dict.keys():
            entity_id_set = self.kg_name_searcher.search_by_keyword(word)
            entity_id_set = valid_id_set.intersection(entity_id_set)

            if len(entity_id_set) == 0:
                continue
            if len(entity_id_set) > self.MAX_CANDIDATE_NODE_NUMBER_FOR_ONE_WORD:
                print("keyword search for %s is %d, too many" % (word, len(entity_id_set)))
                continue

            temp_weight = word_tfidf_weight_dict[word] / len(entity_id_set)

            for entity_id in entity_id_set:
                if entity_id not in entity_weight_dict.keys():
                    entity_weight_dict[entity_id] = 0.0
                entity_weight_dict[entity_id] = entity_weight_dict[entity_id] + temp_weight

        query_vec = self.get_weight_graph_vec(entity_weight_dict)
        return query_vec

    def train_from_doc_collection_with_preprocessor(self, doc_collection: PreprocessMultiFieldDocumentCollection,
                                                    embedding_size=DEFAULT_EMBEDDING_SIZE,
                                                    pretrain_node2vec_path=None,
                                                    graph_data_path=None,
                                                    graph_data=None,
                                                    tfidf_model_path=None,
                                                    tfidf_model=None,
                                                    kg_name_searcher_path=None,
                                                    kg_name_searcher=None,
                                                    doc_sim_model_path=None,
                                                    doc_sim_model_class=None):

        print("start training")
        self.__init_embedding_size(embedding_size)
        self.__init_document_collection(doc_collection)

        corpus_clean_text = []
        preprocess_multi_field_doc_list = doc_collection.get_all_preprocess_document_list()
        for docno, multi_field_doc in enumerate(preprocess_multi_field_doc_list):
            corpus_clean_text.append(multi_field_doc.get_document_text_words())

        print("corpus len=%d" % len(corpus_clean_text))

        print("Dictionary init...")
        self.dict = Dictionary(corpus_clean_text)
        print("Dictionary init complete ")

        print("parse to bow corpus")
        self.corpus = [self.dict.doc2bow(text) for text in corpus_clean_text]
        print("parse to bow corpus complete")

        self.node2vec_model = LoadUtil.load_node2vec_for_doc_collection(doc_collection,
                                                                        pretrain_node2vec_path=pretrain_node2vec_path,
                                                                        embedding_size=embedding_size
                                                                        )

        self.kg_name_searcher = LoadUtil.load_kg_name_searcher(kg_name_searcher=kg_name_searcher,
                                                               kg_name_searcher_path=kg_name_searcher_path)
        self.graph_data = LoadUtil.load_graph_data(graph_data_path=graph_data_path, graph_data=graph_data)

        self.tfidf_model = LoadUtil.load_tfidf_model(tfidf_model=tfidf_model, corpus=self.corpus,
                                                     tfidf_model_path=tfidf_model_path, dict=self.dict)
        self.doc_sim_model_class = doc_sim_model_class
        self.doc_sim_model = doc_sim_model_class.load(doc_sim_model_path)

    def get_weight_graph_vec(self, entity_weight_dict):
        """
        get the average word2vec for list of str
        :param entity_weight_dict:  list of str
        :return: np.array()
        """
        if len(entity_weight_dict.keys()) == 0:
            return self.NP_VECTOR_NOT_EXIST

        valid_entity_id_str = []
        entity_weight_list = []

        for entity_id, w in entity_weight_dict.items():
            entity_id_str = str(entity_id)
            if entity_id_str not in self.node2vec_model.vocab:
                continue
            valid_entity_id_str.append(entity_id_str)
            entity_weight_list.append(w)

        if len(valid_entity_id_str) == 0:
            return self.NP_VECTOR_NOT_EXIST

        entity_vector_list = self.node2vec_model[valid_entity_id_str]
        if entity_vector_list is []:
            return self.NP_VECTOR_NOT_EXIST
        avg_vector = VectorUtil.get_weight_mean_vec(vector_list=entity_vector_list, weight_list=entity_weight_list)

        return avg_vector

    def save(self, model_dir_path):
        """
        save model to the model_dir_path
        :param model_dir_path: the dir to save the model
        :return:
        """
        super().save(model_dir_path)
        self.__init_sub_model_path__()

        self.dict.save(self.dictionary_path)

        print("build dictionary in %s" % self.dictionary_path)

        MmCorpus.serialize(self.corpus_path, self.corpus)
        print("save the corpus to %s" % self.corpus_path)

        print("entity collection saving...")
        self.preprocess_doc_collection.save(self.entity_collection_path)
        print(
            "entity collection finish saving , save to %s, %r" % (
                self.entity_collection_path, self.preprocess_doc_collection))

        print("kg_name_searcher saving...")

        self.kg_name_searcher.save(self.kg_name_searcher_path)
        print("kg_name_searcher saving done...")

        print("node2vec_model saving...")
        self.node2vec_model.save(self.n2v_model_path)
        print("node2vec Training finish , save to %s" % self.n2v_model_path)

        print("graph data saving...")
        self.graph_data.save(self.graph_data_path)
        print("graph data saving done...")

        print("TFIDF model saving...")
        self.tfidf_model.save(self.tfidf_model_path)
        print("TFIDF saving finish , save to %s" % self.tfidf_model_path)

        self.doc_sim_model.save(self.doc_sim_model_path)
        self.save_extra_config()



from pathlib import Path
from unittest import TestCase

from gensim.models.keyedvectors import Word2VecKeyedVectors

from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.models.n2v.svm.avg_n2v import AVGNode2VectorModel
from test.data.definition import ROOT_DIR


class TestAVGNode2VectorModel(TestCase):
    def test_load_name_searcher(self):
        data_dir = Path(ROOT_DIR)
        kg_name_searcher_path = str(data_dir / "JabRef-2.6.v3.namesearcher")
        kg_name_searcher: KGNameSearcher = KGNameSearcher.load(kg_name_searcher_path)
        result = kg_name_searcher.search_by_keyword("pdf")
        print(result)

    def test_n2v(self):
        data_dir = Path(ROOT_DIR)
        node2vec_path = str(data_dir / "JabRef-2.6.v3.weight.node2vec")

        full_node2vec_model = Word2VecKeyedVectors.load(node2vec_path)
        vector = full_node2vec_model["1"]
        result = full_node2vec_model.similar_by_vector(vector=vector, topn=10)
        print(result)
        # todo: doc problem
        result = full_node2vec_model.similar_by_vector(vector=vector, topn=None)
        print(result)
        print(type(result))

    def test_train_from_doc_collection_with_preprocessor(self):
        data_dir = Path(ROOT_DIR)

        model_dir = data_dir / "avg_n2v"
        model_dir.mkdir(exist_ok=True, parents=True)

        doc_collection: PreprocessMultiFieldDocumentCollection = PreprocessMultiFieldDocumentCollection.load(
            str(data_dir / "JabRef-2.6.v3.spacy-pre.sub.dc"))
        embedding_size = 100
        kg_name_searcher_path = str(data_dir / "JabRef-2.6.v3.namesearcher")
        pretrain_node2vec_path = str(data_dir / "JabRef-2.6.v3.weight.node2vec")
        graph_data_path = str(data_dir / "JabRef-2.6.v3.graph")

        AVGNode2VectorModel.train(model_dir_path=str(model_dir),
                                  doc_collection=doc_collection,
                                  embedding_size=embedding_size,
                                  pretrain_node2vec_path=pretrain_node2vec_path,
                                  graph_data_path=graph_data_path,
                                  kg_name_searcher_path=kg_name_searcher_path,
                                  )

    def test_search(self):
        model_dir = Path(ROOT_DIR) / "avg_n2v"
        model_dir.mkdir(exist_ok=True, parents=True)
        query = "download pdf produces unsupported filename"
        model = AVGNode2VectorModel.load(str(model_dir))

        result = model.search(top_num=10, query=query)
        for t in result:
            print(t)
